/*
    Copyright (C) 2011 Collabora Ltd.
      @author George Kiagiadakis <george.kiagiadakis@collabora.co.uk>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef QGLIB_INIT_H
#define QGLIB_INIT_H

#include "global.h"

namespace QGlib {

    /*! Initializes the type system. You must call
     * this function before using any QtGLib API.
     * \note You need to include <QGlib/Init> to use this function.
     */
    QTGLIB_EXPORT void init();

} //namespace QGlib

#endif //QGLIB_INIT_H
